import System.IO

data Color = Black | White deriving (Eq, Show)

type Position = (Int, Int)

data Piece = Piece {
    position :: Position,
    color    :: Color
} deriving (Eq, Show)

type Board = [[Maybe Piece]]

data GameState = GameState {
    board         :: Board,
    currentPlayer :: Color
} deriving (Show)

rows, cols :: Int
rows = 8
cols = 8

-- Original start configurations for isGameOver
blackStart, whiteStart :: [Position]
blackStart = [(0, 0), (0, 1), (0, 2), (0, 3),
              (1, 0), (1, 1), (1, 2),
              (2, 0), (2, 1),
              (3, 0)]

whiteStart = [(4, 7),
              (5, 6), (5, 7),
              (6, 5), (6, 6), (6, 7),
              (7, 4), (7, 5), (7, 6), (7, 7)]

-- Mid-state configuration
blackMid, whiteMid :: [Position]
blackMid = [(0, 0), (0, 3), (1, 3), (1, 5),
            (2, 1), (3, 3), (4, 5),
            (1, 0), (4, 2),
            (5, 5)]

whiteMid = [(2, 2),
            (7, 5), (2, 0),
            (3, 6), (7, 6), (5, 2),
            (5, 3), (5, 6), (6, 6), (7, 3)]

-- Initialize Board with mid-state configuration
initializeBoard :: Board
initializeBoard = [ [ initialPieceAt (r, c) | c <- [0..cols-1] ] | r <- [0..rows-1] ]

initialPieceAt :: Position -> Maybe Piece
initialPieceAt pos
    | pos `elem` blackMid = Just (Piece pos Black)
    | pos `elem` whiteMid = Just (Piece pos White)
    | otherwise           = Nothing

-- Access Board Elements
getPiece :: Board -> Position -> Maybe Piece
getPiece b (r, c) = if inBounds (r, c) then (b !! r) !! c else Nothing

inBounds :: Position -> Bool
inBounds (r, c) = r >= 0 && r < rows && c >= 0 && c < cols

-- Move Pieces
movePiece :: Board -> Position -> Position -> Board
movePiece b from to =
    let piece = getPiece b from
        updatedPiece = fmap (\p -> p { position = to }) piece
        b1 = updateBoard b from Nothing
        b2 = updateBoard b1 to updatedPiece
    in b2

updateBoard :: Board -> Position -> Maybe Piece -> Board
updateBoard b (r, c) val =
    take r b ++
    [take c (b !! r) ++ [val] ++ drop (c + 1) (b !! r)] ++
    drop (r + 1) b

-- Generate Valid Moves
directions :: [Position]
directions = [(dr, dc) | dr <- [-1,0,1], dc <- [-1,0,1], (dr, dc) /= (0, 0)]

getValidMoves :: Board -> Piece -> [Position]
getValidMoves b p =
    let singleMoves = getSingleMoves b p
        jumpMoves = getJumpMoves b (position p) []
    in singleMoves ++ jumpMoves

getSingleMoves :: Board -> Piece -> [Position]
getSingleMoves b p =
    [ (r, c)
    | (dr, dc) <- directions
    , let (r, c) = addPos (position p) (dr, dc)
    , inBounds (r, c)
    , isEmpty b (r, c)
    ]

addPos :: Position -> Position -> Position
addPos (r1, c1) (r2, c2) = (r1 + r2, c1 + c2)

isEmpty :: Board -> Position -> Bool
isEmpty b pos = getPiece b pos == Nothing

getJumpMoves :: Board -> Position -> [Position] -> [Position]
getJumpMoves b pos visited =
    concatMap (\dir -> jumpInDirection b pos dir (pos : visited)) directions

jumpInDirection :: Board -> Position -> Position -> [Position] -> [Position]
jumpInDirection b (r, c) (dr, dc) visited =
    let midPos = (r + dr, c + dc)
        landingPos = (r + 2*dr, c + 2*dc)
    in if inBounds landingPos &&
         not (landingPos `elem` visited) &&
         not (isEmpty b midPos) &&
         isEmpty b landingPos
       then
         let newVisited = landingPos : visited
             furtherJumps = getJumpMoves b landingPos newVisited
         in landingPos : furtherJumps
       else []

-- Check for Game Over (uses original blackStart/whiteStart)
isGameOver :: Board -> Maybe Color
isGameOver b
    | allPiecesInZone b White blackStart = Just White
    | allPiecesInZone b Black whiteStart = Just Black
    | otherwise                          = Nothing

allPiecesInZone :: Board -> Color -> [Position] -> Bool
allPiecesInZone b colorPiece zone =
    let pieces = [ p | row <- b, Just p <- row, color p == colorPiece ]
    in not (null pieces) && all (\p -> position p `elem` zone) pieces

-- Evaluate Board
evaluateBoard :: Board -> Int
evaluateBoard b =
    let whiteScore = sum [ manhattanDistance (position p) (0, 0) | row <- b, Just p <- row, color p == White ]
        blackScore = sum [ manhattanDistance (position p) (7, 7) | row <- b, Just p <- row, color p == Black ]
    in blackScore - whiteScore

manhattanDistance :: Position -> Position -> Int
manhattanDistance (r1, c1) (r2, c2) = abs (r1 - r2) + abs (c1 - c2)

-- Minimax Algorithm with Alpha-Beta Pruning (Sequential)
minimax :: GameState -> Int -> Int -> Int -> Bool -> (Int, GameState)
minimax gameState depth alpha beta maximizingPlayer =
    let b = board gameState
    in case isGameOver b of
        Just winner -> if winner == White then (10000, gameState) else (-10000, gameState)
        Nothing ->
            if depth == 0
            then (evaluateBoard b, gameState)
            else
                let moves = getAllMoves gameState
                    initialEval = if maximizingPlayer then (minBound, gameState) else (maxBound, gameState)
                in alphaBeta moves initialEval alpha beta maximizingPlayer depth

alphaBeta :: [GameState] -> (Int, GameState) -> Int -> Int -> Bool -> Int -> (Int, GameState)
alphaBeta [] bestEval _ _ _ _ = bestEval
alphaBeta (gameState:rest) (bestVal, bestState) alpha beta maximizingPlayer depth =
    let (eval, _) = minimax gameState (depth - 1) alpha beta (not maximizingPlayer)
        (newBestVal, newBestState) =
            if maximizingPlayer
            then if eval > bestVal then (eval, gameState) else (bestVal, bestState)
            else if eval < bestVal then (eval, gameState) else (bestVal, bestState)
        newAlpha = if maximizingPlayer then max alpha eval else alpha
        newBeta  = if not maximizingPlayer then min beta eval else beta
    in if newBeta <= newAlpha
       then (newBestVal, newBestState)  -- Prune remaining moves
       else alphaBeta rest (newBestVal, newBestState) newAlpha newBeta maximizingPlayer depth

getAllMoves :: GameState -> [GameState]
getAllMoves gameState =
    let b = board gameState
        colorPiece = currentPlayer gameState
        pieces = [ p | row <- b, Just p <- row, color p == colorPiece ]
        moves = [ (p, dest) | p <- pieces, dest <- getValidMoves b p ]
        nextPlayer = switchPlayer colorPiece
        gameStates = [ GameState (movePiece b (position p) dest) nextPlayer | (p, dest) <- moves ]
    in gameStates

switchPlayer :: Color -> Color
switchPlayer Black = White
switchPlayer White = Black

-- Command-Line Interface
displayBoard :: Board -> IO ()
displayBoard b = do
    putStrLn "  0 1 2 3 4 5 6 7"
    mapM_ displayRow (zip [0..] b)

displayRow :: (Int, [Maybe Piece]) -> IO ()
displayRow (i, row) = do
    putStr (show i ++ " ")
    putStrLn $ concatMap displayCell row

displayCell :: Maybe Piece -> String
displayCell Nothing                 = ". "
displayCell (Just (Piece _ Black))  = "B "
displayCell (Just (Piece _ White))  = "W "

-- Main: Just one move by the AI (White)
main :: IO ()
main = do
    hSetBuffering stdout NoBuffering
    let initialState = GameState initializeBoard White
    gameLoop initialState

gameLoop :: GameState -> IO ()
gameLoop gameState = do
    displayBoard (board gameState)
    case isGameOver (board gameState) of
        Just winner -> putStrLn $ show winner ++ " wins!"
        Nothing     -> do
            aiTurn gameState

aiTurn :: GameState -> IO ()
aiTurn gameState = do
    putStrLn "AI is thinking..."
    let (eval, newGameState) = minimax gameState 3 (minBound :: Int) (maxBound :: Int) True
    let actualEval = evaluateBoard (board newGameState)
    putStrLn $ "AI evaluated the chosen move with a score of: " ++ show actualEval
    displayBoard (board newGameState)

